// creates images from bullfrog's Buildings Sprites and Map files
// making full map view in BMP format


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bulcommn.h"
#include "xsprblk.h"
#include "xsimmap2.h"
#include "xmapwspr.h"

#ifdef MAIN_XMAPWSPR

int main (int argc, char **argv)
{
    printf("\nBullfrog Engine 2D Maps With Sprites Extractor for Syndicate");
    printf("\n(probably do not work with other Bullfrog games)");
    printf("\n-------------------------------\n");
    if (argc<4)
    {
        printf("Not enought parameters.\n");
        printf("Usage:\n");
        printf("  %s <mapfile> <datfile> <palfile>\n","xmapwspr");
        printf("Output: pic_map.bmp\n");
    	return 1;
    }

    printf("Loading map titles ...");
    MAPTILELIST maptiles;
    int ret=create_map_tiles_rawlist(&maptiles,argv[2],argv[3]);
    if (ret!=0)
        switch (ret)
        {
        case SPRBLK_CANNOT_OPEN_DAT:
             printf("\nError: Cannot open DAT file: %s\n",argv[2]);
             return 2;
        case SPRBLK_CONTENT_BAD_DAT:
             printf("\nError: DAT file content in not right: %s\n",argv[2]);
             return 6;
        case SPRBLK_CANNOT_OPEN_PAL:
             printf("\nError: Cannot open PAL file: %s\n",argv[3]);
             return 3;
        case SPRBLK_CONTENT_BAD_PAL:
             printf("\nError: PAL file content in not right: %s\n",argv[3]);
             return 7;
        default:
             printf("\nError: Map sprites loading returned fail code %d\n",ret);
             return 4;
        }
    printf("Done.\n");

    printf("Loading map data...");
    MAPDATA map;
    ret=create_map_data(&map, argv[1]);
    if (ret!=0)
        switch (ret)
        {
        case 1:
             printf("\nError: Cannot open map file: %s\n",argv[1]);
             return 1;
          case 3:
             printf("\nError: Map file is compressed with RNC, please decompress it\n");
             return 3;
        default:
             printf("\nError: Map data loading returned fail code %d\n",ret);
             return 9;
        }
    printf("Done.\n");

    int scale=1;
    int rgb_mode=0;
    if (scale!=1) rgb_mode=1;
    //All data loaded, now it's time for image generation
    printf("Generating map image...");
    //Full map size in pixels
    long full_map_w=get_tiled_map_full_dimension_w(&map,&maptiles);
    long full_map_h=get_tiled_map_full_dimension_h(&map,&maptiles);
    //if we don't need map endings, we can decrease size a little bit
    long mapwidth=full_map_w/scale;//*4/5;
    long mapheight=full_map_h/scale;//*7/8;
    char *image;
    if (rgb_mode)
        image=malloc(mapwidth*mapheight*3);
    else
        image=malloc(mapwidth*mapheight);
    if (image==NULL)
    {
        printf("\nError: cannot allocate memory for image.\n");
        return 5;
    }
    if (rgb_mode)
        draw_tiled_map_on_buffer_rgb(image,&map,&maptiles,mapwidth,mapheight,full_map_w/2,full_map_h/2,scale);
    else
        draw_tiled_map_on_buffer_idx(image,&map,&maptiles,mapwidth,mapheight,full_map_w/2,full_map_h/2);
    printf("Done.\n");
    //Writting the bitmap
    char fname[20];
  	sprintf (fname, "pic_map.bmp");
    printf("Writing image to %s...",fname);
    if (rgb_mode)
        write_bmp_fn_24b(fname, mapwidth, mapheight,image);
    else
        write_bmp_fn_idx(fname, mapwidth, mapheight, maptiles.palette, image, 0, 1, 2, 4);
    //Finished
    free(image);
    printf("Done.\n");
    free_map_tiles_rawlist(&maptiles);
    free_map_data(&map);
    return 0;
}

#endif

void draw_tiled_map_on_buffer_idx(unsigned char *buffer,MAPDATA *map,MAPTILELIST *tilelist,int width,int height,int wpos,int hpos)
// writes the map to a memory buffer in indexed form
//wpos and hpos are coordinates of the centre of the displayed map
{
/* Map items (x,y) coordinates are as follows:
              /   \
             / 0,0 \
           / \     / \
          /   \   /   \
         / 0,1 \ / 1,0 \
       / \     / \     / \
      /   \   /   \   /   \
     / 0,2 \ / 1,1 \ / 2,0 \
   / \     / \     / \     / \
     .......................
   (and so on)
*/
    //Cleaning the buffer
//    int i;
//    for( i=0; i < (width * height); i++  )
//        buffer[i] = 0;
  //it will be easier to keep tile dimensions here
    int tile_width=64;
    int tile_height=48;
    if (tilelist->num_tiles>0)
    {
         tile_width=tilelist->tiles[0].width;
         tile_height=tilelist->tiles[0].height;
    }
    //Full map size in pixels
    long full_map_w=get_tiled_map_full_dimension_w(map,tilelist);
    long full_map_h=get_tiled_map_full_dimension_h(map,tilelist);
    //Processed map coordinate range
    //TODO: FIX VALUES for partial loading!!!
    int map_start_y=0;
    int map_end_y=map->max_y;
    int map_start_x=0;
    int map_end_x=map->max_x;
    //The change of y coordinate on depth step
    int depth_delta=tile_height/3;
    //Now going through tiles and creating map
    int map_x;
    for (map_x=map_start_x;map_x<map_end_x;map_x++)
    {
        int map_y;
        for (map_y=map_start_y;map_y<map_end_y;map_y++)
        {
            //These equations can be optimized, but it don't make any difference...
            int screen_w = wpos + (width-full_map_w)/2 - (map_y+17)*tile_width/2 + map_x*tile_width/2;
            int screen_h = hpos + height/2-full_map_h  + (map_x+map_y)*tile_height/3;
            int map_idx=((map_y*map->max_x) + map_x)*map->max_z;
            int map_z;
            for (map_z=0;map_z<map->max_z;map_z++)
                {
                int tile_idx=map->data[map_idx + map_z];
                int coord_h=screen_h-((map_z-1)*depth_delta);
                draw_tile_on_buffer_idx(buffer,tilelist,width,height,screen_w,coord_h,tile_idx);
                }
        }
    }
}

void draw_tiled_map_on_buffer_rgb(unsigned char *buffer,MAPDATA *map,MAPTILELIST *tilelist,int width,int height,int wpos,int hpos,int scale)
//Writes the map to a memory buffer in red-green-blue form,
// wpos and hpos are coordinates of the centre of the displayed map.
{
  //it will be easier to keep tile dimensions here
    int tile_width=64;
    int tile_height=48;
    if (tilelist->num_tiles>0)
    {
         tile_width=tilelist->tiles[0].width/scale;
         tile_height=tilelist->tiles[0].height/scale;
    }
    //Full map size in pixels
    long full_map_w=get_tiled_map_full_dimension_w(map,tilelist)/scale;
    long full_map_h=get_tiled_map_full_dimension_h(map,tilelist)/scale;
    //Processed map coordinate range
    //TODO: FIX VALUES for partial loading!!!
    int map_start_y=0;
    int map_end_y=map->max_y;
    int map_start_x=0;
    int map_end_x=map->max_x;
    //The change of y coordinate on depth step
    int depth_delta=tile_height/3;
    //Now going through tiles and creating map
    int map_x;
    for (map_x=map_start_x;map_x<map_end_x;map_x++)
    {
        int map_y;
        for (map_y=map_start_y;map_y<map_end_y;map_y++)
        {
            //These equations can be optimized, but it don't make any difference...
            int screen_w = wpos/scale + (width-full_map_w)/2 + map_x*tile_width/2 - (map_y+17)*tile_width/2;
            int screen_h = hpos/scale + height/2-full_map_h  + (map_x+map_y)*tile_height/3;
            int map_idx=((map_y*map->max_x) + map_x)*map->max_z;
            int map_z;
            for (map_z=0;map_z<map->max_z;map_z++)
                {
                int tile_idx=map->data[map_idx + map_z];
                int coord_h=screen_h-((map_z-1)*depth_delta);
                draw_tile_on_buffer_rgb(buffer,tilelist,width,height,screen_w,coord_h,tile_idx,scale);
                }
        }
    }
}

int draw_tile_on_buffer_idx(unsigned char *buffer,MAPTILELIST *tilelist,int width,int height,int wpos,int hpos,int tile_idx)
//Draws a tile number tile_idx to buffer (whitch has width*height bytes),
// starting (top left corner) at (wpos,hpos).
{
    if (tile_idx >= tilelist->num_tiles) return 1;
    MAPTILE *tile=&(tilelist->tiles[tile_idx]);
    int widx;
    int hidx;
    for (widx=0;widx<tile->width;widx++)
      for (hidx=0;hidx<tile->height;hidx++)
      {
          int buf_h=hpos+hidx;
          int buf_w=wpos+widx;
          if ((buf_h>=0)&&(buf_h<height)&&(buf_w>=0)&&(buf_w<width))
          {
              int buf_idx=buf_h*width + buf_w;
              int til_idx=hidx*tile->width + widx;
              if (tile->alpha[til_idx]<128)
                  buffer[buf_idx]=tile->data[til_idx];
          }
      }
    return 0;
}

int draw_tile_on_buffer_rgb(unsigned char *buffer,MAPTILELIST *tilelist,int width,int height,int wpos,int hpos,int tile_idx,int scale)
//Draws a tile number tile_idx to buffer (whitch has width*height*3 bytes),
// starting (top left corner) at (wpos,hpos).
{
    if (tile_idx >= tilelist->num_tiles) return 1;
    MAPTILE *tile=&(tilelist->tiles[tile_idx]);
    int widx;
    int hidx;
    int scaled_width=tile->width/scale;
    int scaled_height=tile->height/scale;
    for (widx=0;widx<scaled_width;widx++)
      for (hidx=0;hidx<scaled_height;hidx++)
      {
          int buf_h=hpos+hidx;
          int buf_w=wpos+widx;
          if ((buf_h>0)&&(buf_h<height)&&(buf_w>0)&&(buf_w<width))
          {
              int buf_idx=(buf_h*width + buf_w)*3;
              int til_idx=(hidx*tile->width + widx)*scale;
              int red=0;
              int green=0;
              int blue=0;
              int alpha=255;
              //Averaging pixel value if map is scaled
              int num_pixels=scale-1;
              int avg_count=0;
              while (num_pixels>=0)
              {
                  if (widx*scale+num_pixels >= tile->width) continue;
                  int color_idx=3*tile->data[til_idx+num_pixels];
                  alpha=tile->alpha[til_idx+num_pixels];
                  if (alpha<128)
                  {
                      red+=  tilelist->palette[color_idx+2];
                      green+=tilelist->palette[color_idx+1];
                      blue+= tilelist->palette[color_idx];
                      avg_count++;
                  }
                  num_pixels--;
              }
              if (avg_count<1) continue;
              //arithmetic average - dividing by number of points after a sum
              red=4*red/avg_count;
              green=4*green/avg_count;
              blue=4*blue/avg_count;
              //And writing new pixel value
              buffer[buf_idx]=  (unsigned char)( (alpha*buffer[buf_idx]+(255-alpha)*red)    /255);
              buffer[buf_idx+1]=(unsigned char)( (alpha*buffer[buf_idx+1]+(255-alpha)*green)/255);
              buffer[buf_idx+2]=(unsigned char)( (alpha*buffer[buf_idx+2]+(255-alpha)*blue) /255);
          }
      }
    return 0;
}

long get_tiled_map_full_dimension_w(MAPDATA *map,MAPTILELIST *tilelist)
// Return the "real" size of the map in pixels - positions on the map
// are relative to this value. Positions are the same no matter what is
// the current drawing scale.
{
    //tile sizes are required for determining bitmap size
    int tile_width=64;
    if (tilelist->num_tiles>0)
         tile_width=tilelist->tiles[0].width;
    return map->max_x*tile_width*43/50;
}

long get_tiled_map_full_dimension_h(MAPDATA *map,MAPTILELIST *tilelist)
// Return the "real" size of the map in pixels - positions on the map
// are relative to this value. Positions are the same no matter what is
// the current drawing scale.
{
    //tile sizes are required for determining bitmap size
    int tile_height=48;
    if (tilelist->num_tiles>0)
         tile_height=tilelist->tiles[0].height;
    return map->max_y*tile_height*39/50;
}
